import unittest

from pkg_resources import resource_filename

from ._phenomizer import TermPair
from ._phenomizer_io import read_ic_mica_data


class TestPhenomizerIO(unittest.TestCase):

    def setUp(self) -> None:
        self.fpath = resource_filename(__name__, 'test_data/tps.csv.gz')

    def test_read_ic_mica_data(self):
        mica = read_ic_mica_data(self.fpath)
        # We have 50 lines in total, 3 lines are header + comments, leaving 47 MICA entries
        self.assertEqual(len(mica), 47)
        # first entry
        self.assertAlmostEqual(mica[TermPair.of("HP:0006060", "HP:0001180")], 1.829056832, delta=5E-9)
        # last entry
        self.assertAlmostEqual(mica[TermPair.of("HP:0006077", "HP:0001018")], 1.487333401, delta=5E-9)



# Package with methods for calculating semantic similarity.
from ._base import SimilarityKernel, SimilarityMatrixCreator
from ._phenomizer import Phenomizer
from ._phenomizer_io import read_ic_mica_data
from ._dummy import DummySimilarityKernel



import numpy as np

from ._base import SimilarityKernel


class DummySimilarityKernel(SimilarityKernel):
    """A dummy similarity kernel that returns a random similarity value in range [0, 1]."""

    def compute(self, patient_a, patient_b):
        return np.random.uniform(0, 1)



import abc

import numpy as np


class ClusteringAlgorithm(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def cluster(self, similarity_matrix, k_clusters):
        """Cluster the similarity matrix
        :param similarity_matrix: the similarity matrix of the patients to cluster
        :param k_clusters: the number of clusters to form
        """
        pass


class KMeans(ClusteringAlgorithm):

    def cluster(self, similarity_matrix, k_clusters):
        """Cluster the similarity matrix
        :param similarity_matrix: the similarity matrix of the patients to cluster
        :param k_clusters: the number of clusters to form
        """
        # implement clustering here, for stub, just return random assignments
        # TODO - implement or remove
        return np.random.randint(0, k_clusters, similarity_matrix.shape[0])



import re
from typing import Dict, Iterable, Sized, Union

import numpy as np

from ._base import SimilarityKernel

HPO_PATTERN = re.compile(r"HP:(?P<ID>\d{7})")


class TermPair:
    """A class to represent a pair of HPO terms."""

    @staticmethod
    def of(a: str, b: str):
        am = HPO_PATTERN.match(a)
        bm = HPO_PATTERN.match(b)
        if not (am and bm):
            raise ValueError(f"Invalid HPO terms a={a}, b={b}")
        return TermPair(int(am.group("ID")), int(bm.group("ID")))

    def __init__(self, a: int, b: int):
        """Create a TermPair from given HPO term IDs.

        :param a: ID of the first HPO term (e.g. 1234567 for `HP:1234567`)
        :param b: ID of the second HPO term (e.g. 1234567 for `HP:1234567`)
        """
        if a < b:
            self._t1 = a
            self._t2 = b
        else:
            self._t1 = b
            self._t2 = a

    @property
    def t1(self) -> str:
        """first HPO term
        :return:
        """
        return f"HP:{self._t1}"

    @property
    def t2(self) -> str:
        """second HPO term
        :return:
        """
        return f"HP:{self._t2}"

    def __eq__(self, other):
        return other and self.t1 == other.t1 and self.t2 == other.t2

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self._t1, self._t2))

    def __repr__(self):
        return f"TermPair(a={self._id_to_hpo_representation(self._t1)}, b={self._id_to_hpo_representation(self._t1)})"

    @staticmethod
    def _id_to_hpo_representation(term_id: int) -> str:
        """
        Turn the integral ID into an HPO CURIE (e.g. 1234 -> `HP:0001234`).
        """
        return 'HP:' + str(term_id).rjust(7, '0')


class Phenomizer(SimilarityKernel):

    def __init__(self, mica_dict: Dict[TermPair, float]):
        self._mica_dict = mica_dict

    def compute(self, patient_a: Union[Sized, Iterable[str]], patient_b: Union[Sized, Iterable[str]]):
        """" Calculate semantic similarity between patients using the Phenomizer algorithm
        :param patient_a: HPO terms of first individual
        :param patient_b: HPO terms of second individual
        """
        if len(patient_a) == 0 or len(patient_b) == 0:
            return 0.

        similarities = np.zeros(shape=(len(patient_a), len(patient_b)), dtype=float)
        for i, hpoA in enumerate(patient_a):
            for j, hpoB in enumerate(patient_b):
                tp = TermPair.of(hpoA, hpoB)
                similarities[i, j] = self._mica_dict.get(tp, 0.)

        max_a = np.max(similarities, axis=1)
        mean_a = max_a.mean()
        max_b = np.max(similarities, axis=0)
        mean_b = max_b.mean()
        return (mean_a + mean_b) * .5



import abc
from random import sample

import numpy as np


class Perturbation(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def perturb(self, patient_list):
        # here perturb patient_data as part of consensus clustering to obtain stability of clusters
        pass


class NoPerturbation(Perturbation):
    def perturb(self, patient_list):
        """Perturb the input data
        :param patient_list: could be a dataframe with a column of HPO ids as lists, a spark dataframe, phenopackets, etc
        """
        return patient_list


class IncompletePhenotyping(Perturbation):
    def __init__(self, number_of_terms_per_patient):
        """Constructor
        :param number_of_terms_per_patient: number of HPO terms to downsample each individual to
        """
        self._number_of_terms_per_patient = number_of_terms_per_patient

    def perturb(self, patient_list):
        """Perturb the input data
        :param patient_list: could be a dataframe with a column of HPO ids as lists, a spark dataframe, phenopackets, etc
        """
        #execution time of this class for 131072 individuals: 1.3 seconds
        #if we decide on the data structure of patient_list, we can optimize this function by not looping over the lists of lists
        perturbed_data = np.zeros(len(patient_list), dtype=object)
        for i, patient in enumerate(patient_list):
            if len(patient) > self._number_of_terms_per_patient: #only do this when number of terms is larger dan number to downsample to
                perturbed_data[i] = np.random.choice(patient, self._number_of_terms_per_patient, replace=False)
            else:
                perturbed_data[i] = patient
        return perturbed_data


class NoiseAdding(Perturbation):
    def __init__(self, number_of_terms_to_add, nodes):
        """Constructor
        :param number_of_terms_to_add: number of HPO terms to add as random noise to each individual
        """
        self._hpo_terms = nodes
        self._number_of_terms_to_add = number_of_terms_to_add

    def perturb(self, patient_list):
        """Perturb the input data
        :param patient_list: could be a dataframe with a column of HPO ids as lists, a spark dataframe, phenopackets, etc
        """
        #execution time of this class for 131072 individuals: 0.3 seconds (excluding construction of HPO graph)
        selected_hpos = list(np.random.choice(self._hpo_terms, (len(patient_list), self._number_of_terms_to_add)))
        new_patient_list = [a + list(b) for a, b in zip(patient_list, selected_hpos)]
        return new_patient_list




import unittest

from ._phenomizer import Phenomizer, TermPair


class PhenomizerTest(unittest.TestCase):

    def setUp(self) -> None:
        mica_dict = PhenomizerTest._create_mica_dict()
        self.phenomizer = Phenomizer(mica_dict)

    def test_normal_input(self):
        # arachnodactyly and portal hypertension
        patient_a = ["HP:0001166", "HP:0001409"]
        # arachnodactyly, hypertension, and intellectual disability
        patient_b = ["HP:0001166", "HP:0000822", "HP:0001249"]
        similarity = self.phenomizer.compute(patient_a, patient_b)
        self.assertAlmostEqual(similarity, 5.416666666, delta=1E-9)

    def test_empty_returns_zero(self):
        # arachnodactyly and portal hypertension
        patient_a = []
        # arachnodactyly, hypertension, and intellectual disability
        patient_b = []
        similarity = self.phenomizer.compute(patient_a, patient_b)
        self.assertAlmostEqual(similarity, 0., delta=1E-9)

    @staticmethod
    def _create_mica_dict():
        arachnodactyly = "HP:0001166"
        abnormality_of_finger = "HP:0001167"
        hypertension = "HP:0000822"
        portal_hypertension = "HP:0001409"
        # intellectual_disability = "HP:0001249"
        return {
            TermPair.of(arachnodactyly, arachnodactyly): 10.,
            TermPair.of(abnormality_of_finger, abnormality_of_finger): 5.,
            TermPair.of(arachnodactyly, abnormality_of_finger): 5.,

            TermPair.of(hypertension, hypertension): 4.,
            TermPair.of(portal_hypertension, portal_hypertension): 5.,
            TermPair.of(hypertension, portal_hypertension): 3.
        }


class TestTermPair(unittest.TestCase):

    def test_good_input(self):
        tp = TermPair.of("HP:1234567", "HP:9876543")
        self.assertEqual(tp.t1, "HP:1234567")
        self.assertEqual(tp.t2, "HP:9876543")

    def test_input_is_sorted(self):
        tp = TermPair.of("HP:9876543", "HP:1234567")
        self.assertEqual(tp.t1, "HP:1234567")
        self.assertEqual(tp.t2, "HP:9876543")

    def test_invalid_input_raises(self):
        # One or more misformatted inputs
        self.assertRaises(ValueError, TermPair.of, "HP:1234567", "HP:123456")
        self.assertRaises(ValueError, TermPair.of, "HP:123456", "HP:1234567")
        self.assertRaises(ValueError, TermPair.of, "HP:12345", "HP:123456")



import csv
import gzip
from typing import Dict

from ._phenomizer import TermPair


def read_ic_mica_data(fpath: str) -> Dict[TermPair, float]:
    """Read a CSV table with information contents of most informative common ancestors from given `fpath`.

    The file is uncompressed on the fly if the file name ends with `.gz`.
    """
    with _open_file(fpath) as fh:
        comments, header = _parse_header(fh, comment_char='#')
        fieldnames = header.split(",")
        reader = csv.DictReader(fh, fieldnames=fieldnames)

        # Read the lines
        mica = {}
        for row in reader:
            pair = TermPair.of(row['term_a'], row['term_b'])
            mica[pair] = float(row['ic_mica'])

        return mica


def _parse_header(fh, comment_char):
    """Parse header into a list of comments and a header line. As a side effect, the `fh` is set to the position where
    CSV parser can take over and read the records.

    :return: a tuple with a list of comment lines and the header line
    """
    comments = []
    header = None
    for line in fh:
        if line.startswith(comment_char):
            comments.append(line.strip())
        else:
            header = line.strip()
            break
    return comments, header


def _open_file(fpath: str):
    if fpath.endswith(".gz"):
        return gzip.open(fpath, 'rt', newline='')
    else:
        return open(fpath, newline='')



from setuptools import setup, find_namespace_packages

# read requirements/dependencies
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

# read description/README.md
with open("README.md", 'r') as fh:
    long_description = fh.read()

setup(name='c2s2',
      version='0.0.1',
      packages=find_namespace_packages(include=['c2s2.*']),
      install_requires=requirements,

      package_data={'': ['test_data/*']},
      long_description=long_description,
      long_description_content_type='text/markdown',

      author='Lex Dingemans, Daniel Danis',
      author_email='a.dingemans@radboudumc.nl',
      url='https://github.com/monarch-initiative/C2S2',
      description='Consensus clustering for a number of individuals with HPO terms or phenopackets.',
      license='BSD 3',
      keywords='clustering, HPO terms, phenopackets',
      zip_safe=False

      # entry_points={'console_scripts': [
          # 'c2s2 = c2s2.main:main'
      # ]}
      )



import numpy as np
from c2s2.base.cluster import ClusteringAlgorithm
from c2s2.base.perturb import Perturbation
from c2s2.base.semsim import SimilarityMatrixCreator

from ._base import ConsensusClustering


class SimpleConsensusClustering(ConsensusClustering):

    def __init__(self, k_clusters, n_resample,
                 sim_matrix_creator: SimilarityMatrixCreator,
                 perturb_method: Perturbation,
                 clustering_algorithm: ClusteringAlgorithm):
        """Constructor

        :param k_clusters: the upper limit of the clusters to investigate. The limit is included
        :param n_resample: number of times to resample/pertube the dataset
        :param sim_matrix_creator: the similarity matrix creator to calculate the phenotypic similarity matrix
                for patients
        :param perturb_method: the method to resample/perturb the input data for consensus clustering
        :param clustering_algorithm: the algorithm to cluster the patient data, i.e. KMeans, DBScan, etc

        """
        if k_clusters < 2:
            raise Exception("Number of clusters should at least be 2.")

        self.k_clusters = list(range(2, k_clusters + 1))
        self._n_resample = n_resample
        self._perturb_method = perturb_method
        self._clustering_algorithm = clustering_algorithm
        self._sim_matrix = sim_matrix_creator
        self.connectivity_matrix = None
        self.pac = None
        # TODO verify properties of sim_kernel, perturb_method, clustering_algorithm

    def consensus_cluster(self, patient_list):
        # CreateSimilarityMatrix for original patient_list?

        M = np.zeros((len(self.k_clusters), len(patient_list), len(patient_list)), dtype=float)

        for i, k in enumerate(self.k_clusters):
            for h in range(self._n_resample):
                perturbed_patient_list = self._perturb_method.perturb(patient_list)
                similarity_matrix = self._sim_matrix.calculate_matrix(perturbed_patient_list)
                clusters = self._clustering_algorithm.cluster(similarity_matrix, k)
                M[i, :, :] = M[i, :, :] + np.array([clusters == a for a in clusters], dtype=int)
                # update connectivity matrix M, triangular because symmetricas, where Mij = 1 means i and j are in the same cluster (clustering labels can be deleted now)
                # if we're not updating connectivity matrix M, we can also store them all and then collapse/average the Ms
                # early stopping? p-value using likelihood ratio?
        self.connectivity_matrix = M / self._n_resample
        # TODO - thresholds can be hyperparameters
        self.pac = np.mean((self.connectivity_matrix > 0.1) & (self.connectivity_matrix < 0.9), axis=(1, 2))

        # determine best clusters here
        # likelihood ratio test? as epsilon/convergence test
        # calculate some clustering metrics (PAC score?) to compare different fits between k_clusters
        return self



import abc

import numpy as np


class SimilarityKernel(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def compute(self, patient_a, patient_b):
        """
        Calculate semantic similarity between patients.

        :param patient_a: HPO terms of first individual
        :param patient_b: HPO terms of second individual
        """
        pass


class SimilarityMatrixCreator:
    # TODO - possibly remove the entire class if the component looks redundant.

    def __init__(self, similarity_kernel: SimilarityKernel):
        """Constructor
        :param similarity_kernel: kernel to calculate similarity
        """
        self._similarity_kernel = similarity_kernel

    def calculate_matrix(self, patient_list):
        """Calculate the similarity matrix using a kernel function
        :param patient_list: could be a dataframe with a column of HPO ids as lists, a spark dataframe, phenopackets, etc
        """
        # here we need to implement a similarity function, can be phenomizer, or something else, for now, stub returns a random matrix
        similarity_matrix = np.zeros((len(patient_list), len(patient_list)))
        for i in range(len(patient_list)):  # this assumes a list of lists for now
            for j in range(len(patient_list)):  # double loop might not work for spark? --> can def be made more efficient, i.e. instead of loop through full matrix, only calculate for triangle of matrix
                similarity_matrix[i, j] = self._similarity_kernel.compute(patient_list[i], patient_list[j])
        return similarity_matrix

